from typing import Any
from typing import Dict
from typing import List
from typing import Union
from pydantic import BaseModel
from pydantic import Extra
from pydantic import HttpUrl

try:
    from typing import Literal
except ImportError:  # pragma: NO COVER
    from typing_extensions import Literal

from . import settings


class _Base(BaseModel):
    """Base class for questions classes. Main responsibility is renaming
    Python parameter names to Javascript's camelCase convention."""

    class Config:

        fields = {
            "kind": "type",
            "all_rows_required": "isAllRowRequired",
            "expression_format": "format",
            "max_value": "max",
            "min_value": "min",
            "required": "isRequired",
        }

        @classmethod
        def fix_name(cls, name):
            words = name.split("_")
            name = ""
            for index, word in enumerate(words):
                if index == 0:
                    name += word
                else:
                    name += word.capitalize()
            return name

        extra = Extra.allow
        alias_generator = fix_name


class Validator(_Base):
    """Base class for all validators."""

    kind: str
    message: str = "Invalid value"


class TextValidator(Validator):
    """Validator for text input."""

    kind: str = "text"
    max_length: int = 0
    min_length: int = 0
    allow_digits: bool = True


class NumericValidator(Validator):
    """Validator for numeric input."""

    kind: str = "numeric"
    max_value: int = 0
    min_value: int = 0


class EmailValidator(Validator):
    """Checks if a value is a valid email address."""

    kind: str = "email"


class RegexValidator(Validator):
    """Checks if a value matches a regular expression."""

    kind: str = "regex"
    regex: str = ""


class ExpressionValidator(Validator):
    """Checks if a question's answer matches a set of conditions."""

    kind: str = "regex"
    expression: str = ""


class Question(_Base):
    """
    Base question class. All questions share these properties, and all can be
    set by passing them as parameters when instantiating the question.
    """

    kind: str
    name: str = ""
    title: str = ""
    description: str = ""
    required: bool = False
    visible: bool = True
    default_value: str = ""
    correct_answer: str = ""
    visible_if: str = ""
    enable_if: str = ""
    start_with_new_line: bool = True
    value_name: str = ""
    required_if: str = ""
    required_error_text: str = ""
    hide_number: bool = True
    indent: int = 0
    title_location: Literal[settings.TITLE_LOCATIONS] = "default"
    description_location: Literal[settings.DESCRIPTION_LOCATIONS] = "default"
    width: str = ""
    max_width: str = "initial"
    min_width: str = "300px"
    use_display_values_in_title: bool = True
    validators: List[Validator] = []
    extra_js: List[HttpUrl] = []
    extra_css: List[HttpUrl] = []


class TextQuestion(Question):
    """
    A question that uses a text input box. It can handle all HTML5 text input
    types.
    """

    kind: str = "text"
    place_holder: str = ""
    input_type: Literal[settings.TEXT_INPUT_TYPES] = "text"
    max_length: int = -1
    max_value: str = ""
    min_value: str = ""
    size: int = 0
    step: str = ""
    text_update_mode: Literal[settings.TEXT_UPDATE_MODES] = "default"
    input_mask: str = ""
    input_format: str = ""
    prefix: str = ""
    auto_unmask: bool = True
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://unpkg.com/inputmask@5.0.3/dist/inputmask.js",
    ]


class ChoicesQuestion(Question):
    """
    Base class for questions that let the user select one or more options
    from a list of choices. The choices can either be set at question
    instantiation time, or come from a JSON web service at run time.
    """

    kind: str = ""
    col_count: int = 4
    choices: List[Union[str, Dict[str, Union[str, HttpUrl]]]] = []
    choices_by_url: Dict[Literal[settings.CHOICES_BY_URL_KEYS], str] = []
    choices_order: Literal[settings.CHOICE_ORDER_VALUES] = "none"
    choices_enable_if: str = ""
    choices_visible_if: str = ""
    hide_if_choices_empty: bool = True
    has_other: bool = False
    other_text: str = "Other"
    other_error_text: str = ""
    other_place_holder: str = ""


class RadioGroupQuestion(ChoicesQuestion):
    """
    Select one option from a group of radio buttons.
    """

    kind: str = "radiogroup"
    show_clear_button: bool = False


class DropdownQuestion(ChoicesQuestion):
    """
    Select one option from a dropdown menu.
    """

    kind: str = "dropdown"
    choices_max: int = 0
    choices_min: int = 0
    choices_step: int = 1
    options_caption: str = ""
    show_options_caption: bool = True


class CheckboxQuestion(ChoicesQuestion):
    """
    Select one or more options from a group of check boxes.
    """

    kind: str = "checkbox"
    has_none: bool = False
    has_select_all: bool = False
    none_text: str = "None"
    select_all_text: str = ""


class ImagePickerQuestion(ChoicesQuestion):
    """
    Display a group of images and let the user pick one or more.
    """

    kind: str = "imagepicker"
    content_mode: Literal[settings.IMAGE_CONTENT_MODE_VALUES] = "image"
    show_label: bool = False
    image_height: int = 200
    image_width: int = 300
    image_fit: Literal[settings.IMAGE_FIT_VALUES] = "none"
    multi_select: bool = False
    has_other: bool = False
    other_text: str = "Other"
    other_error_text: str = ""
    other_place_holder: str = ""


class BooleanQuestion(Question):
    """
    Get the answer for any two-possibility question. Usually true/false, yes/
    no questions.
    """

    kind: str = "boolean"
    label_true: str = ""
    label_false: str = ""
    show_title: bool = False
    value_true: str = "true"
    value_false: str = "false"


class SignaturePadQuestion(Question):
    """
    Allow the user to "sign" something by drawing their signature. Can be used
    to capture user drawings for other types of questions.
    """

    kind: str = "signaturepad"
    height: int = 200
    width: int = 300
    allow_clear: bool = False


class MultipleTextQuestion(Question):
    """
    A text question with multiple related parts. Displays a text box for each
    part.
    """

    kind: str = "multipletext"
    col_count: int = 2
    items: List[Dict[str, str]] = []
    item_size: int = 0


class CommentQuestion(Question):
    """
    A question that uses a text area to get a multi-line text answer.
    """

    kind: str = "comment"
    rows: int = 3
    cols: int = 50
    max_length: int = -1
    place_holder: str = ""
    text_update_mode: Literal[settings.TEXT_UPDATE_MODES] = "default"


class RatingQuestion(Question):
    """
    A question that lets the user select a number on a fixed scale.
    """

    kind: str = "rating"
    min_rate_description: str = ""
    max_rate_description: str = ""
    rate_max: int = 5
    rate_min: int = 1
    rate_step: int = 1
    rate_values: List[Union[int, Dict[str, Union[int, str]]]] = []


class FileQuestion(Question):
    """
    A question for uploading one or more files.
    """

    kind: str = "file"
    show_preview: bool = True
    allow_multiple: bool = False
    store_data_as_text: bool = True
    image_height: int = 100
    image_width: int = 150
    max_size: int = 0
    accepted_types: str = ""
    allow_images_preview: bool = True
    need_confirm_remove_file: bool = False
    wait_for_upload: bool = True


class MatrixQuestion(Question):
    """
    A question that displays rows of radio buttons and allows the user to
    select a value from one of several columns for each row.
    """

    kind: str = "matrix"
    columns: List[Union[str, Dict[Union[int, str], str]]]
    rows: List[Union[str, Dict[Union[int, str], str]]]
    all_rows_required: bool = False
    cells: Dict[str, Dict[str, str]] = {}
    columns_visible_if: str = ""
    rows_order: Literal[settings.ROW_ORDER_VALUES] = "initial"
    rows_visible_if: str = ""
    show_header: bool = True


class MatrixDropdownQuestion(Question):
    """
    A matrix question that can include other types of input controls, like
    dropdowns and text boxes.
    """

    kind: str = "matrixdropdown"
    columns: List[Dict[Union[int, str], Any]]
    rows: List[Union[str, Dict[Union[int, str], str]]]
    all_rows_required: bool = False
    cells: Dict[str, Dict[str, Any]] = {}
    columns_visible_if: str = ""
    rows_order: Literal[settings.ROW_ORDER_VALUES] = "initial"
    rows_visible_if: str = ""
    show_header: bool = True
    cell_type: Literal[settings.MATRIX_CELL_TYPES] = "dropdown"
    choices: List[Any] = []
    column_col_count: int = 1
    column_layout: Literal[settings.MATRIX_COLUMN_LAYOUTS] = "horizontal"
    column_min_width: str = ""
    horizontal_scroll: bool = False
    options_caption: str = ""
    row_title_width: str = ""
    total_text: str = ""


class MatrixDynamicQuestion(Question):
    """
    A matrix dropdown question that allows the user to add or remove new rows
    from their answer.
    """

    kind: str = "matrixdynamic"
    columns: List[Dict[Union[int, str], Any]]
    rows: List[Dict[Union[int, str], str]] = []
    all_rows_required: bool = False
    cells: Dict[str, Dict[str, Any]] = {}
    columns_visible_if: str = ""
    rows_order: Literal[settings.ROW_ORDER_VALUES] = "initial"
    rows_visible_if: str = ""
    show_header: bool = True
    cell_type: Literal[settings.MATRIX_CELL_TYPES] = "dropdown"
    choices: List[Any] = []
    column_col_count: int = 1
    column_layout: Literal[settings.MATRIX_COLUMN_LAYOUTS] = "horizontal"
    column_min_width: str = ""
    horizontal_scroll: bool = False
    options_caption: str = ""
    add_row_location: Literal[settings.MATRIX_ROW_LOCATIONS] = "default"
    add_row_text: str = ""
    allow_add_rows: bool = True
    allow_remove_rows: bool = True
    confirm_delete: bool = False
    confirm_delete_text: str = ""
    default_row_value: Any = ""
    default_value_from_last_row: bool = False
    key_duplication_error: str = ""
    key_name: str = ""
    max_row_count: int = 100
    min_row_count: int = 1
    remove_row_text: str = ""
    row_count: int = 1


class TagBoxQuestion(DropdownQuestion):
    """
    A question that uses the Select2 tag box widget.
    """

    kind: str = "tagbox"
    select2_config: str = ""
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://cdnjs.cloudflare.com/ajax/libs/select2/4.0.4/js/select2.min.js",
    ]
    extra_css: List[HttpUrl] = [
        "https://cdnjs.cloudflare.com/ajax/libs/select2/4.0.4/css/select2.min.css",
    ]


class JQueryUIDatePickerQuestion(TextQuestion):
    """
    A question that uses the JQueryUI date picker.
    """

    kind: str = "datepicker"
    date_format: str = "mm/dd/yy"
    config: str = ""
    max_date: str = ""
    min_date: str = ""
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://code.jquery.com/ui/1.11.4/jquery-ui.min.js",
    ]
    extra_css: List[HttpUrl] = [
        "https://ajax.googleapis.com/ajax/libs/jqueryui/1.8.18/themes/smoothness/jquery-ui.css",
    ]


class BootstrapDatePickerQuestion(TextQuestion):
    """
    A question that uses the bootstrap date picker.
    """

    kind: str = "bootstrapdatepicker"
    date_format: str = "mm/dd/yy"
    start_date: str = ""
    end_date: str = ""
    today_highlight: bool = True
    week_start: int = 0
    clear_button: bool = False
    auto_close: bool = True
    days_of_week_highlighted: str = ""
    disable_touch_keyboard: bool = True
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://unpkg.com/moment@2.24.0/moment.js",
        "https://cdnjs.cloudflare.com/ajax/libs/bootstrap-datepicker/1.9.0/js/bootstrap-datepicker.js",
    ]
    extra_css: List[HttpUrl] = [
        "https://unpkg.com/bootstrap@3.3.7/dist/css/bootstrap.min.css",
        "https://cdnjs.cloudflare.com/ajax/libs/bootstrap-datepicker/1.9.0/css/bootstrap-datepicker.min.css",
    ]


class Select2Question(DropdownQuestion):
    """
    A question that uses the Select2 dropdown widget.
    """

    render_as: str = "select2"
    select2_config: str = ""
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://cdnjs.cloudflare.com/ajax/libs/select2/4.0.4/js/select2.min.js",
    ]
    extra_css: List[HttpUrl] = [
        "https://cdnjs.cloudflare.com/ajax/libs/select2/4.0.4/css/select2.min.css",
    ]


class BarRatingQuestion(DropdownQuestion):
    """
    A question that uses the JQuery bar rating widget.
    """

    kind: str = "barrating"
    rating_theme: Literal[settings.BAR_RATING_THEMES] = "fontawesome-stars"
    show_values: bool = False
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://unpkg.com/jquery-bar-rating",
    ]
    extra_css: List[HttpUrl] = [
        "https://maxcdn.bootstrapcdn.com/font-awesome/latest/css/font-awesome.min.css",
        "https://unpkg.com/jquery-bar-rating@1.2.2/dist/themes/bars-1to10.css",
        "https://unpkg.com/jquery-bar-rating@1.2.2/dist/themes/bars-movie.css",
        "https://unpkg.com/jquery-bar-rating@1.2.2/dist/themes/bars-pill.css",
        "https://unpkg.com/jquery-bar-rating@1.2.2/dist/themes/bars-reversed.css",
        "https://unpkg.com/jquery-bar-rating@1.2.2/dist/themes/bars-horizontal.css",
        "https://unpkg.com/jquery-bar-rating@1.2.2/dist/themes/fontawesome-stars.css",
        "https://unpkg.com/jquery-bar-rating@1.2.2/dist/themes/css-stars.css",
        "https://unpkg.com/jquery-bar-rating@1.2.2/dist/themes/fontawesome-stars-o.css",
    ]


class SortableJSQuestion(CheckboxQuestion):
    """
    A question that uses the JQuery sortable widget.
    """

    kind: str = "sortablelist"
    empty_text: str = ""
    max_answers_count: int = -1
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://unpkg.com/sortablejs@1.7.0/Sortable.js",
    ]


class NoUISliderQuestion(Question):
    """
    A question that uses the NoUISlider JQuery widget.
    """

    kind: str = "nouislider"
    step: int = 1
    range_min: int = 0
    range_max: int = 100
    pips_mode: str = "positions"
    pips_values: List[int] = [0, 25, 50, 75, 100]
    pips_text: List[Union[int, str]] = [0, 25, 50, 75, 100]
    pips_density: int = 5
    orientation: str = "horizontal"
    direction: str = "ltr"
    tooltips: bool = True
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://unpkg.com/nouislider@9.2.0/distribute/nouislider.js",
        "https://unpkg.com/wnumb@1.1.0",
    ]
    extra_css: List[HttpUrl] = [
        "https://unpkg.com/nouislider@9.2.0/distribute/nouislider.min.css",
    ]


class CKEditorQuestion(Question):
    """
    A question that uses the CKEditor JQuery widget.
    """

    kind: str = "editor"
    height: str = "300px"
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://cdn.ckeditor.com/4.14.1/standard/ckeditor.js",
    ]


class BootstrapSliderQuestion(Question):
    """
    A question that uses the bootstrap slider widget.
    """

    kind: str = "bootstrpslider"
    step: int = 1
    range_min: int = 0
    range_max: int = 100
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://cdnjs.cloudflare.com/ajax/libs/bootstrap-slider/10.0.0/bootstrap-slider.js",
    ]
    extra_css: List[HttpUrl] = [
        "https://unpkg.com/bootstrap@3.3.7/dist/css/bootstrap.min.css",
        "https://cdnjs.cloudflare.com/ajax/libs/bootstrap-slider/10.0.0/css/bootstrap-slider.css",
    ]


class EmotionsRatingQuestion(DropdownQuestion):
    """
    A question that uses the emotion ratings JQuery widget.
    """

    kind: str = "emotionsratings"
    emotions: List[str] = ["angry", "disappointed", "meh", "happy", "inLove"]
    emotion_size: int = 30
    emotions_count: int = 5
    bg_emotion: str = "happy"
    emotion_color: str = "FF0066"
    extra_js: List[HttpUrl] = [
        "https://unpkg.com/jquery@3.5.1/dist/jquery.js",
        "https://unpkg.com/emotion-ratings@2.0.1/dist/emotion-ratings.js",
    ]


class MicrophoneQuestion(Question):
    """
    A question that uses the RecordRTC sound recording widgets.
    """

    kind: str = "microphone"
    extra_js: List[HttpUrl] = [
        "https://www.WebRTC-Experiment.com/RecordRTC.js",
    ]


class HtmlBlock(Question):
    """
    An HTML block that can be embedded in a form.
    """

    kind: str = "html"
    html: str = ""


class ImageBlock(Question):
    """
    An block for inserting an image in a form.
    """

    kind: str = "image"
    image_height: int = 200
    image_width: int = 300
    image_fit: Literal[settings.IMAGE_FIT_VALUES] = "none"
    image_link: HttpUrl
    content_mode: Literal[settings.IMAGE_CONTENT_MODE_VALUES] = "image"


class ExpressionBlock(Question):
    """
    A block that inserts the result of an expression in a form.
    """

    kind: str = "expression"
    expression: str
    currency: str = "USD"
    display_style: Literal[settings.EXPRESSION_DISPLAY_STYLES] = "none"
    expression_format: str = ""
    maximum_fraction_digits: int = -1
    minimum_fraction_digits: int = -1
    use_grouping: bool = True


class PanelBlock(Question):
    """
    A form panel is a group of questions that go together. A form can contain
    any number of panels, even nested.
    """

    kind: str = "panel"
    inner_indent: int = 1
    elements: List[Question] = []
    question_start_index: str = ""
    question_title_location: Literal[settings.TITLE_LOCATIONS] = "default"
    show_number: bool = False
    show_question_numbers: Literal[settings.SHOW_QUESTION_NUMBERS_VALUES] = "default"
    state: Literal[settings.PANEL_STATES] = "default"


class PanelDynamicBlock(Question):
    """
    Dynamic blocks allow users to manually add or remove panels. A form can contain
    any number of dynamic panels, even nested.
    """

    kind: str = "paneldynamic"
    inner_indent: int = 1
    render_mode: Literal[settings.PANEL_RENDER_MODES] = "list"
    panel_count: int = 1
    panel_add_text: str = ""
    panel_remove_text: str = ""
    template_title: str = ""
    template_elements: List[Question] = []
    allow_add_panel: bool = True
    allow_remove_panel: bool = True
    confirm_delete: bool = False
    confirm_delete_text: str = ""
    default_value_from_last_panel: bool = False
    key_duplication_error: str = ""
    key_name: str = ""
    max_panel_count: int = 100
    min_panel_count: int = 1
    panel_add_text: str = ""
    panel_next_text: str = ""
    panel_prev_text: str = ""
    panel_remove_text: str = ""
    panels_state: Literal[settings.PANEL_STATES] = "default"
    show_question_numbers: Literal[settings.SHOW_QUESTION_NUMBERS_VALUES] = "default"
    show_range_in_progress: bool = True
    template_description: str = ""
    template_title_location: Literal[settings.TITLE_LOCATIONS] = "default"


class Page(_Base):
    """
    A group of questions presented as an individual form page. A form can contain
    any number of pages.
    """

    name: str = ""
    title: str = ""
    questions: List[Question] = []
    description: str = ""
    max_time_to_finish: int = 0
    navigation_buttons_visibility: Literal[settings.NAV_BUTTONS_VISIBILITY] = "inherit"
    question_title_location: Literal[settings.TITLE_LOCATIONS] = "default"
    questions_order: Literal[settings.QUESTION_ORDER_VALUES] = "default"


class Survey(_Base):
    """
    The main SurveyJS form container. Questions forms generate a survey
    rendering the questions contained in a form.
    """

    title: str = ""
    pages: List[Page] = []
    calculated_values: List[Any] = []
    check_errors_mode: Literal[settings.CHECK_ERRORS_MODES] = "onNextPage"
    clear_invisible_values: Literal[settings.CLEAR_INVISIBLE_VALUES] = "onComplete"
    completed_before_html: str = ""
    completed_html: str = ""
    completed_html_on_condition: List[Dict[str, str]] = []
    complete_text: str = ""
    cookie_name: str = ""
    description: str = ""
    edit_text: str = ""
    first_page_is_started: bool = False
    focus_first_question_automatic: bool = True
    focus_on_first_error: bool = True
    go_next_page_automatic: bool = False
    loading_html: str = ""
    locale: Literal[settings.LOCALES] = ""
    logo: HttpUrl = ""
    logo_fit: Literal[settings.IMAGE_FIT_VALUES] = "contain"
    logo_height: int = 200
    logo_position: Literal[settings.LOGO_POSITIONS] = "left"
    logo_width: int = 300
    max_others_length: int = 0
    max_text_length: int = 0
    max_time_to_finish: int = 0
    mode: Literal[settings.SURVEY_MODES] = "edit"
    navigate_to_url: HttpUrl = ""
    navigate_to_url_on_condition: List[Dict[str, HttpUrl]] = []
    page_next_text: str = ""
    page_prev_text: str = ""
    preview_text: str = ""
    progress_bar_type: Literal[settings.PROGRESS_BAR_TYPES] = "pages"
    question_description_location: Literal[
        settings.QUESTION_DESCRIPTION_LOCATIONS
    ] = "underTitle"
    question_error_location: Literal[settings.QUESTION_ERROR_LOCATIONS] = "top"
    questions_on_page_mode: Literal[settings.QUESTION_PAGE_MODES] = "standard"
    questions_order: Literal[settings.QUESTION_ORDER_VALUES] = "initial"
    question_start_index: str = ""
    question_title_location: Literal[settings.TITLE_LOCATIONS] = "top"
    question_title_pattern: str = "numTitleRequire"
    question_title_template: str = ""
    required_text: str = "*"
    send_result_on_page_next: bool = False
    show_completed_page: bool = False
    show_navigation_buttons: Literal[settings.NAV_BUTTONS_POSITIONS] = "bottom"
    show_page_numbers: bool = True
    show_page_titles: bool = True
    show_prev_button: bool = True
    show_preview_before_complete: Literal[settings.SHOW_PREVIEW_VALUES] = "noPreview"
    show_progress_bar: Literal[settings.SHOW_PROGRESS_BAR_OPTIONS] = "off"
    show_question_numbers: Literal[settings.PAGE_SHOW_QUESTION_NUMBERS_VALUES] = "on"
    show_timer_panel: Literal[settings.SHOW_TIMER_VALUES] = "none"
    show_timer_panel_mode: Literal[settings.SHOW_TIMER_MODES] = "all"
    show_title: bool = True
    start_survey_text: str = ""
    store_others_as_comment: bool = True
    survey_id: str = ""
    survey_post_id: str = ""
    survey_show_data_saving: bool = True
    text_update_mode: Literal[settings.TEXT_UPDATE_MODES] = "onBlur"
    triggers: List[Any] = []


QUESTION_TYPES = [
    BarRatingQuestion,
    BooleanQuestion,
    BootstrapDatePickerQuestion,
    BootstrapSliderQuestion,
    CheckboxQuestion,
    CKEditorQuestion,
    CommentQuestion,
    DropdownQuestion,
    EmotionsRatingQuestion,
    FileQuestion,
    ImagePickerQuestion,
    JQueryUIDatePickerQuestion,
    MatrixDropdownQuestion,
    MatrixDynamicQuestion,
    MatrixQuestion,
    MicrophoneQuestion,
    MultipleTextQuestion,
    NoUISliderQuestion,
    RadioGroupQuestion,
    RatingQuestion,
    Select2Question,
    SignaturePadQuestion,
    SortableJSQuestion,
    TagBoxQuestion,
    TextQuestion,
]


QUESTION_NAMES_TO_TYPES = {
    "barrating": BarRatingQuestion,
    "boolean": BooleanQuestion,
    "bootstrapdatepicker": BootstrapDatePickerQuestion,
    "bootstrapslider": BootstrapSliderQuestion,
    "checkbox": CheckboxQuestion,
    "editor": CKEditorQuestion,
    "comment": CommentQuestion,
    "dropdown": DropdownQuestion,
    "emotionsratings": EmotionsRatingQuestion,
    "expression": ExpressionBlock,
    "file": FileQuestion,
    "html": HtmlBlock,
    "image": ImageBlock,
    "imagepicker": ImagePickerQuestion,
    "datepicker": JQueryUIDatePickerQuestion,
    "matrixdropdown": MatrixDropdownQuestion,
    "matrixdynamic": MatrixDynamicQuestion,
    "matrix": MatrixQuestion,
    "microphone": MicrophoneQuestion,
    "multipletext": MultipleTextQuestion,
    "nouislider": NoUISliderQuestion,
    "radiogroup": RadioGroupQuestion,
    "rating": RatingQuestion,
    "select2": Select2Question,
    "signaturepad": SignaturePadQuestion,
    "sortablelist": SortableJSQuestion,
    "tagbox": TagBoxQuestion,
    "text": TextQuestion,
}



"""
Console scripts for questions.

 - list_resources PLATFORM THEME [--include-widgets]
     Lists all CSS and JS resources needed to run SurveyJS

 - download_surveyjs PATH PLATFORM THEME
     Downloads all CSS and JS resources needed to run SurveyJS
"""
import click
import requests

from .questions import QUESTION_TYPES
from .settings import SURVEY_JS_CDN
from .templates import get_platform_js_resources
from .templates import get_theme_css_resources


@click.command()
@click.argument("path", type=click.Path(exists=True))
@click.argument("platform")
@click.argument("theme")
def download_surveyjs(path, platform, theme):
    """
    Download SurveyJS files to specified location.

    Supplied path must exist.

    Platforms are: angular, jquery, knockout, react, vue.

    Themes are: default, bootstrap, darkblue, darkrose, modern, orange, stone
                winter, winterstone.
    """
    click.echo()
    js = get_platform_js_resources(platform=platform, resource_url=SURVEY_JS_CDN)
    css = get_theme_css_resources(theme=theme, resource_url=SURVEY_JS_CDN)
    widgets_js = set()
    widgets_css = set()
    for question in QUESTION_TYPES:
        extra_js = question.__fields__.get("extra_js").get_default()
        widgets_js.update(extra_js)
        extra_css = question.__fields__.get("extra_css").get_default()
        widgets_css.update(extra_css)
    for url in js + css + list(widgets_js) + list(widgets_css):
        resource = requests.get(url, allow_redirects=True)
        filename = path + "/" + resource.url.split("/")[-1]
        with open(filename, "wb") as downloaded:
            downloaded.write(resource.content)
            click.echo(f"Downloaded {filename}")
    click.echo()


@click.command()
@click.argument("platform")
@click.argument("theme")
@click.option(
    "--include-widgets", default=False, help="include widget resources", is_flag=True
)
def list_resources(platform, theme, include_widgets):
    """
    List required resources for a platform and theme.

    Platforms are: angular, jquery, knockout, react, vue.

    Themes are: default, bootstrap, darkblue, darkrose, modern, orange, stone
                winter, winterstone.
    """
    click.echo()
    js = get_platform_js_resources(platform=platform, resource_url=SURVEY_JS_CDN)
    css = get_theme_css_resources(theme=theme, resource_url=SURVEY_JS_CDN)
    click.echo("Required Javascript resources:")
    for url in js:
        click.echo(f"    {url}")
    if include_widgets:
        click.echo()
        click.echo("Widget specific Javascript resources:")
        for question in QUESTION_TYPES:
            extra_js = question.__fields__.get("extra_js").get_default()
            if extra_js:
                click.echo()
                click.echo(f"{question.__name__}:")
                for url in extra_js:
                    click.echo(f"    {url}")
        click.echo()
    click.echo()
    click.echo("Required CSS resources:")
    for url in css:
        click.echo(f"    {url}")
    if include_widgets:
        click.echo()
        click.echo("Widget specific CSS resources:")
        for question in QUESTION_TYPES:
            extra_css = question.__fields__.get("extra_css").get_default()
            if extra_css:
                click.echo()
                click.echo(f"{question.__name__}:")
                for url in extra_css:
                    click.echo(f"    {url}")
    click.echo()



import json
import re

from typing import Any
from typing import Dict
from typing import Type

try:
    from typing import Literal
except ImportError:  # pragma: NO COVER
    from typing_extensions import Literal

from .questions import Page
from .questions import PanelBlock
from .questions import PanelDynamicBlock
from .questions import Question
from .questions import QUESTION_NAMES_TO_TYPES
from .questions import Survey
from .settings import INCLUDE_KEYS
from .settings import SURVEY_JS_CDN
from .settings import SURVEY_JS_PLATFORMS
from .settings import SURVEY_JS_THEMES
from .settings import SURVEY_JS_WIDGETS
from .templates import get_form_page
from .templates import get_platform_js_resources
from .templates import get_survey_js
from .templates import get_theme_css_resources
from .validators import call_validator
from .validators import ValidationError


RENAMED_FIELDS = {
    "type": "kind",
    "isAllRowRequired": "all_rows_required",
    "format": "expression_format",
    "max": "max_value",
    "min": "min_value",
    "isRequired": "required",
}


def to_camel_case(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


class Form(object):
    """
    This is the base class used for creating user-defined forms. In addition to
    setting up the form configuration and performing validation, it generates
    the SurveyJS form JSON and keeps track of the required Javascript and CSS
    resources.

    :param name:
        The name of the form. If empty, the class name is used.
    :param action:
        The URL where the form data will be posted. If empty, the
        same URL for the form is used.
    :param html_id:
        The id for the div element that will be used to render the form.
    :param theme:
        The name of the base theme for the form. Default value is 'default'.
    :param platform:
        The JS platform to use for generating the form. Default value is 'jquery'.
    :param resource_url:
        The base URL for the theme resources. If provided,
        Questions will expect to find all resources under this URL. If
        empty, the SurveyJS CDN will be used for all resources.
    :param params:
        Optional list of parameters to be passed to the SurveyJS form object.
    """

    questions_resource_url = SURVEY_JS_CDN

    @classmethod
    def set_resource_url(cls, url: str):
        cls.questions_resource_url = url

    def __init__(
        self,
        name: str = "",
        action: str = "",
        html_id: str = "questions_form",
        theme: Literal[SURVEY_JS_THEMES] = "default",
        platform: Literal[SURVEY_JS_PLATFORMS] = "jquery",
        resource_url: str = None,
        **params,
    ):
        if name == "":
            name = self.__class__.__name__
        self.name = name
        self.action = action
        self.html_id = html_id
        self.theme = theme
        self.platform = platform
        if resource_url is None:
            resource_url = self.questions_resource_url
        self.resource_url = resource_url
        self.params = params
        self._extra_js = []
        self._extra_css = []
        self._form_elements = {}

    def __call__(self, form_data=None):
        return self.render_html(form_data=form_data)

    @classmethod
    def from_json(
        cls,
        form_json: str,
        name: str,
    ):
        """
        Generate a Form class definition from properly formatted JSON data. The
        The generated form class can then be instantiated as needed.

        :param form_json:
            A well formed JSON string with a SurveyJS definition.
        :param name:
            The name of the generated class.

        :Returns:
            A new Python Type that is a subclass of Form.
        """
        NewForm = type(name, (cls,), {})
        form_json = json.loads(form_json)
        elements = form_json.items()
        cls._add_type_elements(NewForm, elements)
        return NewForm

    @classmethod
    def _add_type_elements(cls, NewForm, elements):
        """
        Recursively go through the JSON elements to add all specified form
        elements to the passed in Type.
        """
        for name, element in elements:
            if name == "pages":
                for page in element:
                    page_title = page.get("title", "Page")
                    page_name = page.get("name", page_title)
                    NewPage = type(page_name, (cls,), {})
                    page_items = {}
                    page_params = {}
                    for key, value in page.items():
                        if key in ["questions", "elements"]:
                            page_items[key] = value
                        else:
                            if key in RENAMED_FIELDS:
                                new_key = RENAMED_FIELDS[key]
                            else:
                                new_key = to_camel_case(key)
                            page_params[new_key] = value
                    if "name" in page_params:
                        del page_params["name"]
                    if page_items:
                        cls._add_type_elements(NewPage, page_items.items())
                    form_page = FormPage(NewPage, name=page_name, **page_params)
                    setattr(NewForm, page_name, form_page)
            elif name == "questions" or name == "elements":
                for question_element in element:
                    if question_element["type"] in QUESTION_NAMES_TO_TYPES:
                        question_params = {}
                        for key, value in question_element.items():
                            if key in RENAMED_FIELDS:
                                new_key = RENAMED_FIELDS[key]
                            else:
                                new_key = to_camel_case(key)
                            question_params[new_key] = value
                        new_element = QUESTION_NAMES_TO_TYPES[question_element["type"]](
                            **question_params
                        )
                        setattr(NewForm, new_element.name, new_element)
                    elif question_element["type"] in ["panel", "paneldynamic"]:
                        panel_title = question_element.get("title", "Panel")
                        panel_name = question_element.get("name", panel_title)
                        Panel = type(panel_name, (cls,), {})
                        dynamic = question_element["type"] == "paneldynamic"
                        panel_items = {}
                        panel_params = {}
                        for key, value in question_element.items():
                            if key in ["questions", "elements"]:
                                panel_items[key] = value
                            else:
                                if key in RENAMED_FIELDS:
                                    new_key = RENAMED_FIELDS[key]
                                else:
                                    new_key = to_camel_case(key)
                                panel_params[new_key] = value
                        if "name" in panel_params:
                            del panel_params["name"]
                        if panel_items:
                            cls._add_type_elements(Panel, panel_items.items())
                        form_panel = FormPanel(
                            Panel, name=panel_name, dynamic=dynamic, **panel_params
                        )
                        setattr(NewForm, panel_name, form_panel)
            else:
                NewForm.params[name] = element

    def _construct_survey(self):
        """
        Goes through all the form elements and creates a Survey object, which will
        be used to generate the JSON for initializing SurveyJS. As a side effect,
        populates extra CSS and JS resources. Also keeps a dictionary of all form
        elements, used by validation and update object methods.
        """
        self._extra_js = []
        self._extra_css = []
        self._form_elements = {}
        default_page = Page(name="default")
        survey = Survey(**self.params)
        survey.pages.append(default_page)
        self._add_elements(survey, self, top_level=True)
        # get rid of duplicates
        self._extra_js = list(set(self._extra_js))
        self._extra_css = list(set(self._extra_css))
        if self._extra_js:
            self._extra_js.append(f"{self.resource_url}/{SURVEY_JS_WIDGETS}")
        return survey

    def _add_elements(self, survey, form, top_level=False, container_name="questions"):
        """
        Method to put form elements inside a container. Needs to be recursive so that
        pages and panels are properly nested.
        """
        extra_js = []
        extra_css = []
        for element_name, element in form.__class__.__dict__.items():
            if isinstance(element, (FormPage, FormPanel, Question)):
                name = getattr(element, "name", "")
                if name == "":
                    element.name = element_name
                if isinstance(element, FormPage) and top_level:
                    page = Page(name=name, **element.params)
                    self._add_elements(page, element.form)
                    survey.pages.append(page)
                elif isinstance(element, (FormPage, FormPanel)):
                    container = getattr(survey, container_name)
                    if element.dynamic:
                        panel = PanelDynamicBlock(name=name, **element.params)
                        new_container_name = "template_elements"
                    else:
                        panel = PanelBlock(name=name, **element.params)
                        new_container_name = "elements"
                    self._add_elements(
                        panel, element.form, container_name=new_container_name
                    )
                    container.append(panel)
                else:
                    self._form_elements[element.name] = element
                    if element.extra_js != []:
                        for js in element.extra_js:
                            url = js
                            if self.resource_url != SURVEY_JS_CDN:
                                filename = js.split("/")[-1]
                                url = f"{self.resource_url}/{filename}"
                            if (
                                url not in extra_js
                                and url not in self._extra_js
                                and url not in self.required_js
                            ):
                                extra_js.append(url)
                    if element.extra_css != []:
                        for css in element.extra_css:
                            url = css
                            if self.resource_url != SURVEY_JS_CDN:
                                filename = css.split("/")[-1]
                                url = f"{self.resource_url}/{filename}"
                            if (
                                url not in extra_css
                                and url not in self._extra_css
                                and url not in self.required_css
                            ):
                                extra_css.append(url)
                    if top_level:
                        container = survey.pages[0].questions
                    else:
                        container = getattr(survey, container_name)
                    container.append(element)
        self._extra_js = self._extra_js + extra_js
        self._extra_css = self._extra_css + extra_css

    @property
    def extra_js(self):
        """
        Any extra JS resources required by the form's question types.
        """
        self._construct_survey()
        return self._extra_js

    @property
    def extra_css(self):
        """
        Any extra CSS resources required by the form's question types.
        """
        self._construct_survey()
        return self._extra_css

    @property
    def required_js(self):
        """
        Required JS resources needed to run SurveyJS on chosen platform.
        """
        return get_platform_js_resources(self.platform, self.resource_url)

    @property
    def required_css(self):
        """
        Required CSS resources needed to run SurveyJS on chosen platform.
        """
        return get_theme_css_resources(self.theme, self.resource_url)

    @property
    def js(self):
        """
        Combined JS resources for this form.
        """
        return self.required_js + self.extra_js

    @property
    def css(self):
        """
        Combined CSS resources for this form.
        """
        return self.required_css + self.extra_css

    def to_json(self):
        """
        Convert the form to JSON, in the SurveyJS format.

        :Returns:
            JSON object with the form definition.
        """
        survey = self._construct_survey()
        return survey.json(by_alias=True, include=INCLUDE_KEYS)

    def render_js(self, form_data: Dict[str, Any] = None):
        """
        Generate the SurveyJS initialization code for the chosen platform.

        :param form_data: answers to show on the form for each
            question (for edit forms).

        :Returns:
            String with the generated javascript.
        """
        return get_survey_js(
            form_json=self.to_json(),
            form_data=form_data,
            html_id=self.html_id,
            action=self.action,
            theme=self.theme,
            platform=self.platform,
        )

    def render_html(self, title: str = None, form_data: Dict[str, Any] = None):
        """
        Render a full HTML page showing this form.

        :param title:
            The form title.
        :param form_data:
            answers to show on the form for each question (for edit forms).

        :Returns:
            String with the generated HTML.
        """
        if title is None:
            title = self.params.get("title", self.name)
        if form_data is None:
            form_data = {}
        survey_js = self.render_js(form_data=form_data)
        return get_form_page(
            title=title,
            html_id=self.html_id,
            platform=self.platform,
            survey_js=survey_js,
            js_resources=self.js,
            css_resources=self.css,
        )

    def validate(self, form_data: Dict[str, Any], set_errors: bool = False):
        """
        Server side validation mimics what client side validation should do. This
        means that any validation errors here are due to form data being sent from
        outside the SurveyJS form, possibly by directly posting the data to the form.
        Questions keeps track of the errors, even though the UI will show them anyway.
        Validation returns False if at least one validator doesn't pass.

        :param form_data:
            A dictionary-like object with the form data to be validated.
        :param set_errors:
            set to :data:`True` to add an `__errors__` key to the
            form data dictionary, containing the validation errors.

        :Returns:
            :data:`True` if the validation passes, :data:`False` otherwise.
        """
        validated = True
        errors = []
        self._construct_survey()
        for name, element in self._form_elements.items():
            value = form_data.get(name)
            if value is None and element.required:
                errors.append({"question": name, "message": "An answer is required"})
                validated = False
            for validator in element.validators:
                if not call_validator(validator, value, form_data):
                    errors.append({"question": name, "message": validator.message})
                    validated = False
        if set_errors:
            form_data["__errors__"] = errors
        return validated

    def update_object(self, obj: Any, form_data: Dict[str, Any]):
        """
        Utility method to set an object's attributes with data obtained from a form.
        This method validates the data before setting the object's attributes.

        :param obj:
            The object to set attributes on.

        :param form_data:
            A dictionary-like object with the form data to be validated.

        :Raises:
            questions.validators.ValidationError if validation does not pass.
        """
        if self.validate(form_data):
            # call to validate sets the form elements beforehand
            for name in self._form_elements.keys():
                setattr(obj, name, form_data[name])
        else:
            raise ValidationError


class FormPage(object):
    """
    Represents an individual page from a multi-page form.

    :param form:
        A subclass of questions.Form (not an instance). The form
        to be shown in its own page.
    :param name:
        The name of the form.
    :param params:
        Optional list of parameters to be passed to the SurveyJS page object.
    """

    def __init__(self, form: Type[Form], name: str = "", **params):
        self.form = form()
        if name == "":
            name = self.form.name
        self.name = name
        self.dynamic = False
        self.params = params


class FormPanel(object):
    """
    A panel is a set of fields that go together. It can be used for visual
    separation, or as a dynamically added group of fields for complex questions.

    :param form:
        A subclass of questions.Form (not an instance). The form
        to be shown in its own page.
    :param name:
        The name of the form.
    :param dynamic:
        Set to :data:`True` if the panel will be used as a template for adding
         or removing groups of questions.
    :param params:
        Optional list of parameters to be passed to the SurveyJS panel object.
    """

    def __init__(
        self,
        form: Type[Form],
        name: str = "",
        dynamic: bool = False,
        **params,
    ):
        self.form = form()
        if name == "":
            name = self.form.name
        self.name = name
        self.dynamic = dynamic
        self.params = params



import json
from typing import Any
from typing import Dict
from typing import List

from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import select_autoescape

from .settings import BOOTSTRAP_URL
from .settings import SUGGESTED_JS_BY_PLATFORM
from .settings import SURVEY_JS_CDN


# Initialize Jinja environment
env = Environment(
    loader=PackageLoader("questions", "templates"),
    autoescape=select_autoescape(["html", "xml"]),
)


def _render_template(
    kind: str = "js",
    platform: str = "jquery",
    **context_data: Dict,
):
    filename = f"survey_{kind}.{platform}.jinja"
    template = env.get_template(filename)
    return template.render(**context_data)


def get_platform_js_resources(
    platform: str = "jquery",
    resource_url: str = SURVEY_JS_CDN,
):
    """
    Get the list of suggested JS resources for a platform. if not using the
    CDN, only the main SurveyJS JS file is returned.

    :param platform:
        The name of the JS platform.
    :param resource_url:
        The URL where all SurveyJS resources are located.

    :Returns:
        The list of resource URLs.
    """
    survey_js = f"{resource_url}/survey.{platform}.min.js"
    platform_js = []
    for js in SUGGESTED_JS_BY_PLATFORM[platform]:
        if resource_url == SURVEY_JS_CDN:
            platform_js.append(js)
        else:
            filename = js.split("/")[-1]
            platform_js.append(f"{resource_url}/{filename}")
    platform_js.append(survey_js)
    return platform_js


def get_theme_css_resources(
    theme: str = "default",
    resource_url: str = SURVEY_JS_CDN,
):
    """
    Get the list of suggested CSS resources for a theme. if not using the
    CDN, only the main SurveyJS CSS file is returned.

    :param theme:
        The name of the CSS theme.
    :param resource_url:
        The URL where all SurveyJS resources are located.

    :Returns:
        The list of resource URLs.
    """
    if theme == "bootstrap" and resource_url == SURVEY_JS_CDN:
        return [BOOTSTRAP_URL]
    elif theme == "bootstrap":
        return [f"{resource_url}/bootstrap.min.css"]
    name = "survey"
    if theme == "modern":
        name = "modern"
    return [f"{resource_url}/{name}.css"]


def get_survey_js(
    form_json: str = "",
    form_data: Dict[str, Any] = None,
    html_id: str = "questions_form",
    action: str = "",
    theme: str = "default",
    platform: str = "jquery",
):
    """
    Get the SurveyJS initialization script and form definition.

    :param form_json:
        The JSON generated from the questions form.
    :param form_data:
        Any form data to set on the rendered form.
    :param html_id:
        The HTML id of the form placeholder.
    :param action:
        The URL where the submitted form data will be posted.
    :param theme:
        The name of the SurveyJS theme to use.
    :param platform:
        The name of the supported SurveyJS platform to use.

    :Returns:
        The rendered JS as string.
    """
    if form_data is None:
        form_data = {}
    data = json.dumps(form_data)
    return _render_template(
        kind="js",
        platform=platform,
        theme=theme,
        json=form_json,
        data=data,
        action=action,
        html_id=html_id,
    )


def get_form_page(
    title: str = "",
    html_id: str = "questions_form",
    platform: str = "jquery",
    survey_js: str = "",
    js_resources: List = None,
    css_resources: List = None,
):
    """
    Generate a standalone SurveyJS HTML page.

    :param title:
        The form title to display.
    :param html_id:
        The HTML id of the form placeholder.
    :param platform:
        The name of the supported SurveyJS platform to use.
    :param survey_js:
        The generated JS to put in the form.
    :param js_resources:
        The list of JS resources to add to the HTMl head.
    :param css_resources:
        The list of CSS resources to add to the HTMl head.

    :Returns:
        The rendered HTML as string.
    """
    resources = ""
    if js_resources is None:
        js_resources = []
    if css_resources is None:
        css_resources = []
    for resource in js_resources:
        resources += f'<script src="{resource}"></script>\n'
    for resource in css_resources:
        resources += f'<link href="{resource}" type="text/css" rel="stylesheet" />\n'
    return _render_template(
        kind="html",
        platform=platform,
        title=title,
        resources=resources,
        html_id=html_id,
        survey_js=survey_js,
    )



#!/usr/bin/env python

"""Tests for `cli` package."""

import os

from click.testing import CliRunner

from questions import cli


def test_command_list_resources():
    runner = CliRunner()
    result = runner.invoke(cli.list_resources, ["jquery", "default"])
    assert result.exit_code == 0
    assert "Required Javascript resources" in result.output
    assert "Required CSS resources" in result.output


def test_command_list_resources_include_widgets():
    runner = CliRunner()
    result = runner.invoke(
        cli.list_resources, ["jquery", "default", "--include-widgets"]
    )
    assert result.exit_code == 0
    assert "Required Javascript resources" in result.output
    assert "Required CSS resources" in result.output
    assert "Widget specific Javascript resources" in result.output
    assert "Widget specific CSS resources" in result.output


def test_command_download_surveyjs():
    runner = CliRunner()
    with runner.isolated_filesystem():
        os.mkdir("_tmp_")
        result = runner.invoke(cli.download_surveyjs, ["_tmp_", "jquery", "default"])
        assert result.exit_code == 0
        assert "Downloaded " in result.output



#!/usr/bin/env python

"""Tests for `validators` package."""

from questions import validators
from questions import EmailValidator
from questions import ExpressionValidator
from questions import NumericValidator
from questions import RegexValidator
from questions import TextValidator


def test_text_validator_bad_min_length():
    validator = TextValidator(min_length=7)
    assert validators.text_validator(validator, "hello", {}) is False
    assert validators.text_validator(validator, "goodbye", {}) is True


def test_text_validator_bad_max_length():
    validator = TextValidator(max_length=4)
    assert validators.text_validator(validator, "hello", {}) is False
    assert validators.text_validator(validator, "Bye", {}) is True


def test_text_validator_allow_digits():
    validator = TextValidator(allow_digits=False)
    assert validators.text_validator(validator, "h4ll0", {}) is False
    validator = TextValidator(allow_digits=True)
    assert validators.text_validator(validator, "h4ll0", {}) is True


def test_numeric_validator_bad_min_value():
    validator = NumericValidator(min_value=7)
    assert validators.numeric_validator(validator, 5, {}) is False
    assert validators.numeric_validator(validator, 45, {}) is True


def test_numeric_validator_bad_max_value():
    validator = NumericValidator(max_value=42)
    assert validators.numeric_validator(validator, 99, {}) is False
    assert validators.numeric_validator(validator, 22, {}) is True


def test_email_validator():
    validator = EmailValidator()
    assert validators.email_validator(validator, "my.bad.address.com", {}) is False
    assert validators.email_validator(validator, "my@true.email.com", {}) is True


def test_regex_validator():
    validator = RegexValidator(regex="[A-Z][a-z]+-[0-9][0-9]")
    assert validators.regex_validator(validator, "anything", {}) is False
    assert validators.regex_validator(validator, "Abc-55", {}) is True


def test_expression_validator():
    validator = ExpressionValidator(expression="{var2} notempty and {var3} > 5")
    form_data = {"var1": "something", "var2": "", "var3": 1}
    assert validators.expression_validator(validator, "anything", form_data) is False
    assert validators.expression_validator(validator, "anything", {}) is False
    form_data = {"var1": "something", "var2": "other", "var3": 8}
    assert validators.expression_validator(validator, "anything", form_data) is True
    validator = ExpressionValidator(expression="")
    assert validators.expression_validator(validator, "anything", form_data) is True



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['Click>=7.0',
                'email_validator',
                'Jinja2',
                'pydantic',
                'requests',
                'simpleeval',
                ]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3.6', ]

setup(
    author="Carlos de la Guardia",
    author_email='cguardia@yahoo.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    description="Questions is a form library that uses the power of SurveyJS for the UI.",
    entry_points={
        'console_scripts': [
            'download_surveyjs=questions.cli:download_surveyjs',
            'list_resources=questions.cli:list_resources',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='questions',
    name='questions',
    packages=find_packages(include=['questions', 'questions.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/cguardia/questions',
    version='0.5.0a1',
    zip_safe=False,
)
